"""Ensure model output dtypes are consistent with model inputs."""

from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from pycontrails import Flight, GeoVectorDataset, MetDataset
from pycontrails.core.met import MetDataArray
from pycontrails.models.aircraft_performance import AircraftPerformance
from pycontrails.models.cocip import Cocip
from pycontrails.models.humidity_scaling import ConstantHumidityScaling
from pycontrails.models.issr import ISSR
from pycontrails.models.sac import SAC
from pycontrails.physics import units
from tests import BADA3_PATH, BADA4_PATH, BADA_AVAILABLE


@pytest.fixture
def met(met_ecmwf_pl_path: str) -> MetDataset:
    """Create a MetDataset for testing."""
    ds1 = xr.open_dataset(met_ecmwf_pl_path)

    # shift time and concatenate to create a new dataset
    ds2 = ds1.copy()
    ds2["time"] = ds2["time"] + np.timedelta64(1, "h")
    ds = xr.concat([ds1, ds2], dim="time")
    for v in ds.data_vars:
        assert ds[v].dtype == "float32"

    met = MetDataset(ds)
    assert met.shape == (15, 8, 3, 4)
    return met


@pytest.fixture(params=["float32", "float64"])
def mda(met: MetDataset, request) -> MetDataArray:
    """Create a MetDataArray for testing."""
    mda = met["t"]
    mda.data = mda.data.astype(request.param)
    return mda


@pytest.mark.parametrize("vector_dtype", ["float32", "float64", "int32", "int64"])
@pytest.mark.parametrize("method", ["linear", "nearest"])
def test_interpolation_out_dtype(mda: MetDataArray, method: str, vector_dtype: str):
    """Ensure interpolation output dtype is consistent with grid dtype."""

    t0 = mda.data.time.values[0] + np.timedelta64(30, "m")
    vector = GeoVectorDataset(
        longitude=np.array([1, 2, 3], dtype=vector_dtype),
        latitude=np.array([1, 2, 3], dtype=vector_dtype),
        level=np.array([251, 252, 253], dtype=vector_dtype),
        time=[t0, t0, t0],
    )
    out = vector.intersect_met(mda, bounds_error=True, method=method)
    assert out.size == 3
    assert out.dtype == mda.data.dtype


def test_unit_conversion():
    """Confirm basic unit conversion maintains the array dtype."""
    level = np.array([1, 2, 3], dtype="float32")
    assert level.dtype == "float32"
    altitude = units.pl_to_ft(level)
    assert altitude.dtype == "float32"
    altitude_ft = units.m_to_ft(altitude)
    assert altitude_ft.dtype == "float32"
    level_ft = units.ft_to_pl(altitude_ft)
    assert level_ft.dtype == "float32"


@pytest.mark.parametrize("model_class", [ISSR, SAC])
@pytest.mark.parametrize("method", ["linear", "nearest"])
def test_issr_sac_grid_output(met_issr: MetDataset, model_class: type, method: str):
    """Confirm ISSR and SAC gridded output is float32 when met input is float32.

    Something in dask threading is causing this test to occasionally hang. As a
    workaround, we load the met data into memory before passing it to the model.
    """
    met_issr.data.load()

    assert all(v == "float32" for v in met_issr.data.dtypes.values())

    model = model_class(
        met_issr,
        interpolation_method=method,
        humidity_scaling=ConstantHumidityScaling(),
    )
    out = model.eval()
    assert isinstance(out, MetDataArray)
    assert out.data.dtype == "float32"
    assert out.shape == met_issr.shape


@pytest.fixture(scope="module")
def random_vector(met_issr: MetDataset):
    """Create a random vector with coordinates mostly matching `met_issr`."""
    # Build random vector
    # Purposely going out of bounds from met
    rng = np.random.default_rng(22777)
    longitude = rng.uniform(-180, 180, 1000).astype("float32")
    latitude = rng.uniform(-90, 90, 1000).astype("float32")
    level = rng.uniform(200, 300, 1000).astype("float32")

    t0 = met_issr.data.time.values[0]
    time = t0 + rng.integers(0, 60, 1000) * np.timedelta64(1, "m")

    return GeoVectorDataset(
        longitude=longitude,
        latitude=latitude,
        level=level,
        time=time,
    )


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("model_class", [ISSR, SAC])
@pytest.mark.parametrize("method", ["linear", "nearest"])
def test_issr_sac_vector_output(
    met_issr: MetDataset,
    model_class: type,
    method: str,
    random_vector: GeoVectorDataset,
):
    """Confirm ISSR and SAC vector output is float32 when source and met input is float32."""
    # Run the model
    model = model_class(
        met_issr,
        interpolation_method=method,
        humidity_scaling=ConstantHumidityScaling(),
    )
    out = model.eval(random_vector)
    assert isinstance(out, GeoVectorDataset)
    assert out.size == 1000
    assert out[model.name].dtype == "float32"
    assert out["air_temperature"].dtype == "float32"
    assert out["specific_humidity"].dtype == "float32"

    # Confirm interpolation behaves as expected
    in_bounds = random_vector.coords_intersect_met(met_issr)
    interp_non_nan = np.isfinite(out["air_temperature"])
    np.testing.assert_array_equal(in_bounds, interp_non_nan)


@pytest.mark.skipif(not BADA4_PATH.is_dir(), reason="BADA4 not available")
@pytest.mark.parametrize("drop_aircraft_performance", [False, True])
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_cocip(
    met_cocip1: MetDataset,
    rad_cocip1: MetDataset,
    flight_cocip1: Flight,
    drop_aircraft_performance: bool,
    bada_model: AircraftPerformance,
):
    """Confirm CoCiP maintains float32 precision."""
    # Convert flight_cocip1 dtypes to float32
    # Grab all the keys up front to avoid mutating while iterating
    keys = [k for k in flight_cocip1 if k != "time"]
    for k in keys:
        flight_cocip1.update({k: flight_cocip1[k].astype("float32")})

    if drop_aircraft_performance:
        for k in set(keys).difference(["longitude", "latitude", "altitude"]):
            del flight_cocip1[k]
        flight_cocip1.attrs.update(aircraft_type="B737")

    # Run the model
    cocip = Cocip(
        met=met_cocip1,
        rad=rad_cocip1,
        process_emissions=drop_aircraft_performance,
        humidity_scaling=ConstantHumidityScaling(),
        aircraft_performance=bada_model,
    )
    out = cocip.eval(flight_cocip1)

    # Check that the output is float32 or less with a few exceptions
    exclude = {"time", "flight_id", "waypoint", "formation_time", "age", "contrail_age"}
    vectors = [
        cocip._sac_flight,
        cocip._downwash_flight,
        cocip._downwash_contrail,
        out,
        *cocip.contrail_list,
    ]
    for vector in vectors:
        for k in vector:
            if k in exclude:
                continue
            arr = vector[k]
            assert np.can_cast(arr.dtype, "float32"), k


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("bada_priority", [3, 4])
def test_cocip_grid(met_cocip1: MetDataset, rad_cocip1: MetDataset, bada_priority: int):
    """Confirm `CocipGrid` maintains float32 precision."""
    CocipGrid = pytest.importorskip("pycontrails.models.cocipgrid").CocipGrid

    # Keep max_age small to avoid blowing out of bounds
    # And keep dt_integration small to make the evolution interesting
    params = {
        "max_age": np.timedelta64(2, "h"),
        "bada_priority": bada_priority,
        "dt_integration": np.timedelta64(10, "m"),
        "humidity_scaling": ConstantHumidityScaling(),
    }
    if bada_priority == 3:
        params["bada3_path"] = BADA3_PATH
    elif bada_priority == 4:
        params["bada4_path"] = BADA4_PATH
    cg = CocipGrid(met=met_cocip1, rad=rad_cocip1, params=params)

    # Create float32 source
    source = cg.create_source(
        level=250,
        time=cg.met.data.time.values[0],
        longitude=cg.met.data.longitude.values,
        latitude=cg.met.data.latitude.values,
    )
    assert source.shape == (16, 8, 1, 1)

    out = cg.eval(source)
    assert out["ef_per_m"].data.dtype == "float32"
    assert out.attrs["bada_model"] == f"BADA{bada_priority}"
    assert out.attrs["max_age"] == "2 hours"
    assert out.attrs["dt_integration"] == "10 minutes"
    assert out.attrs["pycontrails_version"].startswith("0.")

    # Not much persistence, but we can pin them
    # And importantly, they are the same with BADA3 or BADA4
    nonzero = np.flatnonzero(out["ef_per_m"].data.values.ravel())
    np.testing.assert_array_equal(nonzero, [38, 46, 52, 53, 54, 61, 62, 70])



"""Interface to `pycontrails-cirium` extension."""

from __future__ import annotations

try:
    from pycontrails_cirium import Cirium

    __all__ = ["Cirium"]

except ImportError as e:
    raise ImportError(
        "Failed to import `pycontrails-cirium` extension. Install with `pip install"
        ' "pycontrails-cirium @ git+ssh://git@github.com/contrailcirrus/pycontrails-cirium.git"`'
    ) from e



"""Ice super-saturated regions (ISSR)."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, overload

import numpy as np

import pycontrails
from pycontrails.core.flight import Flight
from pycontrails.core.met import MetDataArray, MetDataset
from pycontrails.core.met_var import AirTemperature, SpecificHumidity
from pycontrails.core.models import Model, ModelParams
from pycontrails.core.vector import GeoVectorDataset
from pycontrails.models.humidity_scaling import HumidityScaling
from pycontrails.physics import constants, thermo
from pycontrails.utils.types import ArrayLike, apply_nan_mask_to_arraylike


@dataclass
class ISSRParams(ModelParams):
    """Default ISSR model parameters."""

    #: RHI Threshold
    rhi_threshold: float = 1.0

    #: Humidity scaling
    humidity_scaling: HumidityScaling | None = None


class ISSR(Model):
    """Ice super-saturated regions over a :class:`Flight` trajectory or :class:`MetDataset` grid.

    This model calculates points where the relative humidity over ice is greater than 1.

    Parameters
    ----------
    met : MetDataset
        Dataset containing "air_temperature" and "specific_humidity" variables

    Examples
    --------
    >>> from datetime import datetime
    >>> from pycontrails.datalib.ecmwf import ERA5
    >>> from pycontrails.models.issr import ISSR
    >>> from pycontrails.models.humidity_scaling import ConstantHumidityScaling

    >>> # Get met data
    >>> time = datetime(2022, 3, 1, 0), datetime(2022, 3, 1, 2)
    >>> variables = ["air_temperature", "specific_humidity"]
    >>> pressure_levels = [200, 250, 300]
    >>> era5 = ERA5(time, variables, pressure_levels)
    >>> met = era5.open_metdataset(xr_kwargs=dict(parallel=False))

    >>> # Instantiate and run model
    >>> scaling = ConstantHumidityScaling(rhi_adj=0.98)
    >>> model = ISSR(met, humidity_scaling=scaling)
    >>> issr = model.eval()
    >>> issr.proportion  # Get proportion of values with ice supersaturation
    0.114140917963

    >>> # Run with a lower threshold
    >>> issr2 = ISSR(met, rhi_threshold=0.95, humidity_scaling=scaling).eval()
    >>> issr2.proportion
    0.146647
    """

    name = "issr"
    long_name = "Ice super-saturated regions"
    met_variables = AirTemperature, SpecificHumidity
    default_params = ISSRParams

    @overload
    def eval(self, source: Flight, **params: Any) -> Flight:
        ...

    @overload
    def eval(self, source: GeoVectorDataset, **params: Any) -> GeoVectorDataset:
        ...

    @overload
    def eval(self, source: MetDataset | None = None, **params: Any) -> MetDataArray:
        ...

    def eval(
        self, source: GeoVectorDataset | Flight | MetDataset | None = None, **params: Any
    ) -> GeoVectorDataset | Flight | MetDataArray:
        """Evaluate ice super-saturated regions along flight trajectory or on meteorology grid.

        .. versionchanged:: 0.27.0

            Humidity scaling now handled automatically. This is controlled by
            model parameter ``humidity_scaling``.

        Parameters
        ----------
        source : GeoVectorDataset | Flight | MetDataset | None, optional
            Input GeoVectorDataset or Flight.
            If None, evaluates at the :attr:`met` grid points.
        **params : Any
            Overwrite model parameters before eval

        Returns
        -------
        GeoVectorDataset | Flight | MetDataArray
            Returns 1 in ISSR, 0 everywhere else.
            Returns `np.nan` if interpolating outside meteorology grid.

        Raises
        ------
        NotImplementedError
            Raises if input ``source`` is not supported.
        """

        self.update_params(params)
        self.set_source(source)

        if isinstance(self.source, GeoVectorDataset):
            self.downselect_met()
            self.source.setdefault("air_pressure", self.source.air_pressure)

        scale_humidity = (self.params["humidity_scaling"] is not None) and (
            "specific_humidity" not in self.source
        )
        self.set_source_met()

        # apply humidity scaling, warn if no scaling is provided for ECMWF data
        if scale_humidity:
            self.params["humidity_scaling"].eval(self.source, copy_source=False)

        issr_ = issr(
            air_temperature=self.source.data["air_temperature"],
            specific_humidity=self.source.data["specific_humidity"],
            air_pressure=self.source.data["air_pressure"],
            rhi=self.source.data.get("rhi", None),  # if rhi already known, pass it in
            rhi_threshold=self.params["rhi_threshold"],
        )
        self.source["issr"] = issr_

        # Tag output with additional attrs when source is MetDataset
        if isinstance(self.source, MetDataset):
            attrs: dict[str, Any] = {
                "description": self.long_name,
                "pycontrails_version": pycontrails.__version__,
            }
            if scale_humidity:
                for k, v in self.params["humidity_scaling"].description.items():
                    attrs[f"humidity_scaling_{k}"] = v
            if self.met is not None:
                attrs["met_source"] = self.met.attrs.get("met_source", "unknown")

            self.source["issr"].data.attrs.update(attrs)
            return self.source["issr"]

        # In GeoVectorDataset case, return source
        return self.source


def issr(
    air_temperature: ArrayLike,
    specific_humidity: ArrayLike | None = None,
    air_pressure: ArrayLike | None = None,
    rhi: ArrayLike | None = None,
    rhi_threshold: float = 1.0,
) -> ArrayLike:
    r"""Calculate ice super-saturated regions.

    Regions where the atmospheric relative humidity over ice is greater than 1.

    Parameters ``air_temperature``, ``specific_humidity``, ``air_pressure``,
    and ``rhi`` must have compatible shapes when defined.

    Either ``specific_humidity`` and ``air_pressure`` must both be provided, or
    ``rhi`` must be provided.

    Parameters
    ----------
    air_temperature : ArrayLike
        A sequence or array of temperature values, :math:`[K]`.
    specific_humidity : ArrayLike | None
        A sequence or array of specific humidity values, [:math:`kg_{H_{2}O} \ kg_{moist air}`]
        None by default.
    air_pressure : ArrayLike | None
        A sequence or array of atmospheric pressure values, [:math:`Pa`]. None by default.
    rhi : ArrayLike | None, optional
        A sequence of array of RHi values, if already known. If not provided, this function
        will compute RHi from `air_temperature`, `specific_humidity`, and `air_pressure`.
        None by default.
    rhi_threshold : float, optional
        Relative humidity over ice threshold for determining ISSR state

    Returns
    -------
    ArrayLike
        ISSR state of each point indexed by the parameters.
    """
    if rhi is None:
        if specific_humidity is None or air_pressure is None:
            raise TypeError(
                "If 'rhi' is not specified, both 'specific_humidity' "
                "and 'air_pressure' must be provided."
            )
        rhi = thermo.rhi(specific_humidity, air_temperature, air_pressure)

    # store nan values to refill after casting
    nan_mask = np.isnan(rhi)

    # compute issr as int
    sufficiently_cold = air_temperature < -constants.absolute_zero
    sufficiently_humid = rhi > rhi_threshold

    issr_ = (sufficiently_cold & sufficiently_humid).astype(rhi.dtype)

    return apply_nan_mask_to_arraylike(issr_, nan_mask)



"""Test thermo formulas as needed for SAC calculations."""


import numpy as np
import pytest

from pycontrails import JetA, VectorDataset
from pycontrails.models.sac import T_sat_liquid, rh_critical_sac, slope_mixing_line
from pycontrails.physics import thermo


@pytest.fixture(scope="module")
def data():
    rng = np.random.default_rng(4321)
    n = 100000

    v = VectorDataset()
    v["air_pressure"] = rng.uniform(5000, 50000, n)
    v["specific_humidity"] = rng.uniform(0, 0.003, n)
    v["air_temperature"] = rng.uniform(150, 250, n)
    return v


@pytest.mark.parametrize("engine_efficiency", [0.2, 0.3, 0.4, 0.5, 0.6])
def test_slope_mixing_line(data: VectorDataset, engine_efficiency: float):
    """Test the range of the slope_mixing_line formula."""
    fuel = JetA()
    G = slope_mixing_line(
        data["specific_humidity"], data["air_pressure"], engine_efficiency, fuel.ei_h2o, fuel.q_fuel
    )

    # For many calculations, we evaluate log(G - 0.053)
    # Consequently, we want G to be bounded above this constant
    # We use a small buffer to quantify this
    buffer = 0.2
    assert np.all(G > 0.053 + buffer)


@pytest.mark.parametrize("e_sat", [thermo.e_sat_ice, thermo.e_sat_liquid])
def test_e_sat_increasing(e_sat):
    """Check that thermo.e_sat_ice, thermo.e_sat_liquid are increasing and positive."""
    T = np.linspace(150, 350, 10000)
    e_sat_ = e_sat(T)
    assert np.all(e_sat_ > 0)
    assert np.all(np.diff(e_sat_) > 0)


@pytest.mark.parametrize("engine_efficiency", [0.2, 0.4, 0.6])
def test_rh_crit(data: VectorDataset, engine_efficiency: float):
    """Confirm that `rh_critical_sac` takes values in [0, 1]."""
    fuel = JetA()
    G = slope_mixing_line(
        data["specific_humidity"], data["air_pressure"], engine_efficiency, fuel.ei_h2o, fuel.q_fuel
    )
    T_sat_liquid_ = T_sat_liquid(G)
    rh_crit = rh_critical_sac(data["air_temperature"], T_sat_liquid_, G)

    # All values nonnegative
    assert np.all(rh_crit >= 0)

    # Some values are infinity, indicating no SAC possible
    is_finite = np.isfinite(rh_crit)
    assert np.all(rh_crit[~is_finite] == np.inf)

    # Finite values are below 1
    assert np.all(rh_crit[is_finite] <= 1)



"""Test cache module."""

from __future__ import annotations

import os
import pathlib
import random
from datetime import datetime

import pytest
from google.cloud import storage

from pycontrails import DiskCacheStore, GCPCacheStore
from pycontrails.core import cache

# see if GCP credentials exist
try:
    storage.Client()
except Exception:
    gcp_credentials = False
else:
    gcp_credentials = True

try:
    import requests

    r = requests.get("https://github.com", timeout=10)
except Exception:
    offline = True
else:
    offline = False

############
# Disk Cache Store
############

DISK_CACHE_DIR = cache._get_user_cache_dir()


class TestDiskCacheStore:
    def test_cache_dir_exists(self) -> None:
        """Test if default cache exists and is a directory."""
        _cache = DiskCacheStore()
        _dir = pathlib.Path(_cache.cache_dir)
        assert _dir.exists()
        assert _dir.is_dir()

        # init with pathlib
        _cache = DiskCacheStore(cache_dir=pathlib.Path(f"{DISK_CACHE_DIR}/test"))
        assert pathlib.Path(_cache.cache_dir).is_dir()
        assert _cache.cache_dir == str(pathlib.Path(_cache.cache_dir))

    def test_cache_bad_path(self) -> None:
        _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test")

        # return False for random string
        disk_path = "".join(random.choice(["x", "y"]) for _ in range(20))
        assert not _cache.exists(disk_path)

    @pytest.mark.skipif(not pathlib.Path("README.md").is_file(), reason="No README.md")
    def test_cache_size(self) -> None:
        _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=True)
        _cache.clear()
        size = _cache.size
        assert isinstance(size, float) and size == 0

        _cache.put("README.md")
        size = _cache.size
        assert isinstance(size, float) and size > 0

        assert _cache.listdir() == ["README.md"]

        # cleanup
        _cache.clear()

    def test_path(self) -> None:
        _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=True)

        disk_path = _cache.path("")
        assert "pycontrails" in disk_path
        assert pathlib.Path(disk_path).is_dir()

    def test_exists(self) -> None:
        _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=True)

        assert not _cache.exists("cache/path4.nc")
        disk_path = _cache.path("cache/path4.nc")
        assert _cache.exists("cache")
        open(disk_path, "w").close()
        assert _cache.exists("cache/path4.nc")
        assert _cache.exists(disk_path)

    def test_cache_allow_clear(self) -> None:
        _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=False)
        with pytest.raises(RuntimeError):
            _cache.clear()

        _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=True)
        disk_path = _cache.path("path.nc")
        open(disk_path, "w").close()
        _cache.clear()
        assert not _cache.exists("path.nc")

    @pytest.mark.skipif(not pathlib.Path("README.md").is_file(), reason="No README.md")
    def test_put_get(self) -> None:
        _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=True)
        _cache.clear()

        # put from filename
        assert not _cache.exists("README.md")
        _cache.put("README.md")
        assert _cache.exists("README.md")

        _cache.put("README.md", "anotherfile.nc")
        assert _cache.exists("anotherfile.nc")

        # file doesn't exist
        with pytest.raises(FileNotFoundError, match="No file found"):
            _cache.put("test2.nc")

        # get from filename
        _path = _cache.get("README.md")
        assert _path == _cache.path("README.md")

        # clean up
        _cache.clear()

    @pytest.mark.skipif(not pathlib.Path("README.md").is_file(), reason="No README.md")
    def test_put_multiple(self) -> None:
        _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=True)
        _cache.clear()

        # put from filenames
        assert not _cache.exists("README2.md")
        _cache.put_multiple(["README.md", ".gitignore"], ["README2.md", ".gitignore2"])
        assert _cache.exists("README2.md")
        assert _cache.exists(".gitignore2")
        _cache.clear()

        # put from pathlibs
        assert not _cache.exists("README2.md")
        _cache.put_multiple(
            [pathlib.Path("README.md"), pathlib.Path(".gitignore")], ["README2.md", ".gitignore2"]
        )
        assert _cache.exists("README2.md")
        assert _cache.exists(".gitignore2")

        # file doesn't exist
        with pytest.raises(FileNotFoundError, match="No file found"):
            _cache.put_multiple(["README2.md", ".gitignore"], ["README2.md", ".gitignore2"])

        # clean up
        _cache.clear()


############
# GCS Cache Store
############
BUCKET = "contrails-301217-unit-test"
CACHE_DIR = datetime.now().isoformat().replace(":", "-")


@pytest.mark.skipif(offline, reason="offline")
@pytest.mark.skipif(not gcp_credentials, reason="No GCS credentials")
class TestGCPCacheStore:
    def test_cache_init(self) -> None:
        # test if default cache exists and is a directory
        _cache = GCPCacheStore(bucket=BUCKET)
        assert _cache.bucket == BUCKET
        assert _cache.cache_dir == (
            f'{os.getenv("PYCONTRAILS_CACHE_DIR")}/'
            if "PYCONTRAILS_CACHE_DIR" in os.environ
            else ""
        )
        assert _cache.read_only is True
        assert _cache.allow_clear is False
        assert ".gcp" in _cache._disk_cache.cache_dir
        assert BUCKET in _cache._disk_cache.cache_dir

        # test no bucket
        with pytest.raises(ValueError, match="Parameter `bucket` not specified"):
            GCPCacheStore()

        # parse gs:// URI
        _cache = GCPCacheStore(bucket=BUCKET, cache_dir=f"gs://{BUCKET}/{CACHE_DIR}/")
        assert _cache.bucket == BUCKET
        assert _cache.cache_dir == f"{CACHE_DIR}/"

        # parse gs:// URI when bucket not specified
        _cache = GCPCacheStore(cache_dir=f"gs://{BUCKET}/{CACHE_DIR}/")
        assert _cache.bucket == BUCKET
        assert _cache.cache_dir == f"{CACHE_DIR}/"

        # conflicting bucket raises an error
        with pytest.raises(ValueError, match="conflicting bucket names"):
            GCPCacheStore(bucket="xyz", cache_dir=f"gs://{BUCKET}/{CACHE_DIR}/")

        # raise https:// urls
        with pytest.raises(ValueError, match="only specify base object path"):
            GCPCacheStore(bucket=BUCKET, cache_dir="https://mybucket/")

        # append /
        _cache = GCPCacheStore(bucket=BUCKET, cache_dir=f"{CACHE_DIR}")
        assert _cache.cache_dir == f"{CACHE_DIR}/"

        # test custom disk cache
        _disk_cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/{CACHE_DIR}")
        _cache = GCPCacheStore(bucket=BUCKET, disk_cache=_disk_cache)
        assert _cache.bucket == BUCKET
        assert (
            ".gcp" not in _cache._disk_cache.cache_dir
            and f"{CACHE_DIR}" in _cache._disk_cache.cache_dir
        )

    def test_cache_size(self) -> None:
        _cache = GCPCacheStore(
            bucket=BUCKET, cache_dir=f"{CACHE_DIR}/", allow_clear=True, read_only=False
        )
        _cache._dangerous_clear(confirm=True)
        size = _cache.size
        assert isinstance(size, float) and size == 0

        _cache.put("README.md")
        size = _cache.size
        assert isinstance(size, float) and size > 0

    def test_cache_allow_clear(self) -> None:
        """Only allowed through _dangerous_clear."""
        _cache = GCPCacheStore(bucket=BUCKET, cache_dir=f"{CACHE_DIR}/", allow_clear=False)
        with pytest.raises(RuntimeError, match="not allowed to be cleared"):
            _cache._dangerous_clear(confirm=True)

        _cache = GCPCacheStore(
            bucket=BUCKET, cache_dir=f"{CACHE_DIR}/", allow_clear=True, read_only=False
        )
        _cache.put("README.md")
        assert _cache.exists("README.md")

        with pytest.raises(RuntimeError):
            _cache._dangerous_clear(confirm=False)

        _cache._dangerous_clear(confirm=True)
        assert not _cache.exists("README.md")
        assert _cache.size == 0

    # Download and upload logic is slightly different with show_progress=True
    # Run a simple smoke test to ensure get and put work as expected
    @pytest.mark.parametrize("show_progress", [False, True])
    def test_put_get(self, show_progress) -> None:
        _cache = GCPCacheStore(
            bucket=BUCKET,
            cache_dir=f"{CACHE_DIR}/",
            allow_clear=True,
            read_only=False,
            show_progress=show_progress,
        )
        _cache._dangerous_clear(confirm=True)

        # put from filename
        assert not _cache.exists("README.md")
        _cache.put("README.md")
        assert _cache.exists("README.md")

        _cache.put("README.md", "anotherfile.nc")
        assert _cache.exists("anotherfile.nc")

        # file doesn't exist
        with pytest.raises(FileNotFoundError, match="No file found"):
            _cache.put("test2.nc")

        # get from filename
        _path = _cache.get("README.md")
        assert _path == _cache._disk_cache.path("README.md")

        # read only
        _cache = GCPCacheStore(
            bucket=BUCKET, cache_dir=f"{CACHE_DIR}/", allow_clear=True, read_only=False
        )
        _cache.put("README.md", "anotherfile.nc")
        _cache = GCPCacheStore(
            bucket=BUCKET, cache_dir=f"{CACHE_DIR}/", allow_clear=True, read_only=True
        )
        with pytest.raises(RuntimeError, match="read only"):
            _cache.put("README.md", "anotherfile2.nc")
        _path = _cache.get("anotherfile.nc")
        assert "anotherfile.nc" in _path

        # clean up
        _cache._dangerous_clear(confirm=True)



"""
Build assets for second set of cocip unit tests.

Built from static data in `/tests/regression/cocip-fortran/inputs`.
See `/tests/regression/cocip-fortran/README.md` for instructions to download inputs.

Run this file from the terminal to generate assets:

```python
$ python cocip-met2.py
```

"""

from __future__ import annotations

import pathlib

import numpy as np
import pandas as pd

from pycontrails import DiskCacheStore, Flight
from pycontrails.datalib.ecmwf import ERA5
from pycontrails.models import Cocip
from pycontrails.physics import units

# use separate disk cache
disk_cache = DiskCacheStore(cache_dir="cache")


# from pycontrails.models.humidity_scaling import ConstantHumidityScaling


# setup
flight_input_filename = "flight1.csv"

input_path = pathlib.Path("../regression/cocip-fortran/inputs")
output_path = pathlib.Path("../unit/static")


## ----
# Load Flight
# ----

# load flight waypoints
df_flight = pd.read_csv(input_path / flight_input_filename)

# constant properties along the length of the flight
attrs = {
    "flight_id": flight_input_filename,
    "aircraft_type": df_flight["ICAO Aircraft Type"].values[0],
    "wingspan": df_flight["Wingspan (m)"].values[0],
}

# convert UTC timestamp to np.datetime64
df_flight["time"] = pd.to_datetime(df_flight["UTC time"], origin="unix", unit="s")

# get altitude in m
df_flight["altitude"] = units.ft_to_m(df_flight["Altitude (feet)"])

# rename a few columns for compatibility with `Flight` requirements
df_flight = df_flight.rename(
    columns={
        "Longitude (degrees)": "longitude",
        "Latitude (degrees)": "latitude",
        "True airspeed (m s-1)": "true_airspeed",
        "Mach Number": "mach_number",
        "Aircraft mass (kg)": "aircraft_mass",
        "Fuel mass flow rate (kg s-1)": "fuel_flow",
        "Overall propulsion efficiency": "engine_efficiency",
        "nvPM number emissions index (kg-1)": "nvpm_ei_n",
    }
)

# clean up a few columns before building Flight class
df_flight = df_flight.drop(
    columns=["ICAO Aircraft Type", "Wingspan (m)", "UTC time", "Altitude (feet)"]
)

fl = Flight(data=df_flight, attrs=attrs)


# get met dims from Flight
time = (
    pd.to_datetime(fl["time"][0]).floor("H"),
    pd.to_datetime(fl["time"][-1]).ceil("H") + pd.Timedelta("5H"),
)
pressure_levels = [
    400,
    350,
    300,
    250,
    225,
    200,
    175,
    150,
]  # select pressure levels

# load remote met - only necessary when not using regression test inputs
# era5pl = ERA5(
#     time=time,
#     variables=Cocip.met_variables,
#     pressure_levels=pressure_levels,
#     cachestore=disk_cache,
# )
# era5sl = ERA5(time=time, variables=Cocip.rad_variables, cachestore=disk_cache)

# load met locally from `input_path`
filenames = [t.strftime("%Y-%m-%dT%H") for t in pd.date_range(time[0], time[1], freq="1H")]

met_filepaths = [str(input_path / "met" / f"{f}.nc") for f in filenames]
rad_filepaths = [str(input_path / "rad" / f"{f}.nc") for f in filenames]


era5pl = ERA5(
    time=time, variables=Cocip.met_variables, pressure_levels=pressure_levels, paths=met_filepaths
)
era5sl = ERA5(time=time, variables=Cocip.rad_variables, paths=rad_filepaths)

# create `MetDataset` from sources
met = era5pl.open_metdataset(xr_kwargs=dict(parallel=False))
rad = era5sl.open_metdataset(xr_kwargs=dict(parallel=False))


params = {
    "downselect_met": True,
    "process_emissions": False,
    "max_age": np.timedelta64(4, "h"),
    # "interpolation_fill_value": 0.0,
    "verbose_outputs": True,
    # "humidity_scaling": ConstantHumidityScaling(rhi_adj=1)
}

cocip = Cocip(met=met, rad=rad, params=params)
fl_out = cocip.eval(source=fl)


## Unit test output
df_flight["flight_id"] = "test2"
df_flight["wingspan"] = 65
df_flight.to_csv(output_path / "flight-cocip2.csv", index=False)


met_test = cocip.met.data
rad_test = cocip.rad.data

step = 8
met_test = met_test.reindex(
    latitude=met_test["latitude"].values[0::step],
    longitude=met_test["longitude"].values[0::step],
    level=[200, 225, 300],
)
rad_test = rad_test.reindex(
    latitude=rad_test["latitude"].values[0::step], longitude=rad_test["longitude"].values[0::step]
)

# outputs
met_test.to_netcdf(output_path / "met-era5-cocip2.nc")
rad_test.to_netcdf(output_path / "rad-era5-cocip2.nc")


print(f"Static test assets written to {output_path}")



"""ECMWF Parameter Support.

Sourced from the ECMWF Parameter DB:
    https://apps.ecmwf.int/codes/grib/param-db
"""

from __future__ import annotations

from pycontrails.core import met_var
from pycontrails.core.met import MetVariable

RelativeVorticity = MetVariable(
    short_name="vo",
    standard_name="atmosphere_upward_relative_vorticity",
    long_name="Vorticity (relative)",
    units="s**-1",
    level_type="isobaricInhPa",
    grib1_id=43,
    ecmwf_id=138,
    grib2_id=(0, 2, 12),
    description=(
        "Atmosphere upward relative vorticity is the vertical component of the 3D air vorticity"
        " vector. The vertical component arises from horizontal velocity only. 'Relative' in this"
        " context means the vorticity of the air relative to the rotating solid earth reference"
        " frame, i.e. excluding the Earth's own rotation. In contrast, the quantity with standard"
        " name atmosphere_upward_absolute_vorticity includes the Earth's rotation. 'Upward'"
        " indicates a vector component which is positive when directed upward (negative downward)."
        " A positive value of atmosphere_upward_relative_vorticity indicates anticlockwise rotation"
        " when viewed from above."
    ),
)

PotentialVorticity = MetVariable(
    short_name="pv",
    standard_name="potential_vorticity",
    long_name="Potential vorticity (K m^2 / kg s)",
    units="K m**2 kg**-1 s**-1",
    level_type="isobaricInhPa",
    grib1_id=128,
    ecmwf_id=60,
    grib2_id=(0, 2, 14),
    amip="pvu",
    description=(
        "Potential vorticity is a measure of the capacity for air to rotate in the atmosphere.If we"
        " ignore the effects of heating and friction, potential vorticity is conserved following an"
        " air parcel.It is used to look for places where large wind storms are likely to originate"
        " and develop.Potential vorticity increases strongly above the tropopause and therefore, it"
        " can also be used in studiesrelated to the stratosphere and stratosphere-troposphere"
        " exchanges. Large wind storms develop when a columnof air in the atmosphere starts to"
        " rotate. Potential vorticity is calculated from the wind, temperature andpressure across a"
        " column of air in the atmosphere."
    ),
)

CloudAreaFractionInLayer = MetVariable(
    short_name="cc",
    standard_name="fraction_of_cloud_cover",
    long_name="Cloud area fraction in atmosphere layer",
    ecmwf_id=248,
    level_type="isobaricInhPa",
    grib2_id=(0, 6, 32),
    units="[0 - 1]",
    amip="cl",
    description=(
        "This parameter is the proportion of a grid box covered by cloud (liquid or ice) at a"
        " specific pressure level."
    ),
)

SpecificCloudLiquidWaterContent = MetVariable(
    short_name="clwc",
    standard_name="specific_cloud_liquid_water_content",
    long_name="Specific cloud liquid water content",
    units="kg kg**-1",
    level_type="isobaricInhPa",
    ecmwf_id=246,
    grib2_id=(0, 1, 83),
    description=(
        "This parameter is the mass of cloud liquid water droplets per kilogram of the total mass"
        " of moist air. The 'total mass of moist air' is the sum of the dry air, water vapour,"
        " cloud liquid, cloud ice, rain and falling snow. This parameter represents the average"
        " value for a grid box."
    ),
)


SpecificCloudIceWaterContent = MetVariable(
    short_name="ciwc",
    standard_name="specific_cloud_ice_water_content",
    long_name="Specific cloud ice water content",
    units="kg kg**-1",
    level_type="isobaricInhPa",
    ecmwf_id=247,
    grib2_id=(0, 1, 84),
    description=(
        "This parameter is the mass of cloud ice particles per kilogram of the total mass of moist"
        " air. The 'total mass of moist air' is the sum of the dry air, water vapour, cloud liquid,"
        " cloud ice, rain and falling snow. This parameter represents the average value for a grid"
        " box."
    ),
)

# Override units and description on Relative humidity
RelativeHumidity = MetVariable(
    short_name=met_var.RelativeHumidity.short_name,
    standard_name=met_var.RelativeHumidity.standard_name,
    long_name=met_var.RelativeHumidity.long_name,
    units="%",
    level_type=met_var.RelativeHumidity.level_type,
    ecmwf_id=met_var.RelativeHumidity.ecmwf_id,
    grib2_id=met_var.RelativeHumidity.grib2_id,
    description=(
        "This parameter is the water vapour pressure as a percentage of the value at which the air"
        " becomes saturated "
        "(the point at which water vapour begins to condense into liquid water or deposition into"
        " ice)."
        "For temperatures over 0°C (273.15 K) it is calculated for saturation over water. "
        "At temperatures below -23°C it is calculated for saturation over ice. "
        "Between -23°C and 0°C this parameter is calculated by interpolating between the ice and"
        " water values using a quadratic function."
        "See https://www.ecmwf.int/sites/default/files/elibrary/2016/17117-part-iv-physical-processes.pdf#subsection.7.4.2"  # noqa: E501
    ),
)

TOAIncidentSolarRadiation = MetVariable(
    short_name="tisr",
    standard_name="toa_incident_solar_radiation",
    long_name="Top of atmosphere incident shortwave radiation",
    units="J m**-2",
    level_type="nominalTop",
    ecmwf_id=212,
    grib2_id=(192, 128, 212),  # reference ECMWF
    description="Top of atmosphere incident solar radiation. Accumulated field.",
)

TopNetSolarRadiation = MetVariable(
    short_name="tsr",
    standard_name="top_net_solar_radiation",
    long_name="Top of atmosphere net solar (shortwave) radiation",
    units="J m**-2",
    level_type="nominalTop",
    ecmwf_id=178,
    grib2_id=(0, 4, 1),
    description=(
        "This parameter is the incoming solar radiation (also known as shortwave radiation) "
        "minus the outgoing solar radiation at the top of the atmosphere. "
        "It is the amount of radiation passing through a horizontal plane. "
        "The incoming solar radiation is the amount received from the Sun. "
        "The outgoing solar radiation is the amount reflected and scattered by the Earth's"
        " atmosphere and surface"
        "See https://www.ecmwf.int/sites/default/files/elibrary/2015/18490-radiation-quantities-ecmwf-model-and-mars.pdf"  # noqa: E501
    ),
)

TopNetThermalRadiation = MetVariable(
    short_name="ttr",
    standard_name="top_net_thermal_radiation",
    long_name="Top of atmosphere net thermal (longwave) radiation",
    units="J m**-2",
    level_type="nominalTop",
    ecmwf_id=179,
    grib2_id=(0, 5, 5),
    description=(
        "The thermal (also known as terrestrial or longwave) "
        "radiation emitted to space at the top of the atmosphere is commonly known as the Outgoing"
        " Longwave Radiation (OLR). "
        "The top net thermal radiation (this parameter) is equal to the negative of OLR."
        "See https://www.ecmwf.int/sites/default/files/elibrary/2015/18490-radiation-quantities-ecmwf-model-and-mars.pdf"  # noqa: E501
    ),
)

SurfaceSolarDownwardRadiation = MetVariable(
    short_name="ssrd",
    standard_name="surface_solar_downward_radiation",
    long_name="Surface Solar Downward Radiation",
    units="J m**-2",
    level_type="surface",
    ecmwf_id=169,
    grib2_id=(0, 4, 7),
    description=(
        "This parameter is the amount of solar radiation (also known as shortwave radiation) that"
        " reaches a horizontal plane at the surface of the Earth. This parameter comprises both"
        " direct and diffuse solar radiation."
    ),
)

CloudAreaFraction = MetVariable(
    short_name="tcc",
    standard_name="total_cloud_cover",
    long_name="Cloud area fraction (total)",
    level_type="surface",
    grib1_id=71,
    ecmwf_id=164,
    grib2_id=(192, 128, 164),  # reference ECMWF
    units="[0 - 1]",
    amip="clt",
    description=(
        "This parameter is the proportion of a grid box covered by cloud (liquid or ice) for a"
        " whole atmosphere column."
    ),
)

PRESSURE_LEVEL_VARIABLES = [
    met_var.AirTemperature,
    met_var.SpecificHumidity,
    met_var.Geopotential,
    met_var.EastwardWind,
    met_var.NorthwardWind,
    met_var.VerticalVelocity,
    RelativeHumidity,
    RelativeVorticity,
    CloudAreaFractionInLayer,
    SpecificCloudIceWaterContent,
    SpecificCloudLiquidWaterContent,
    PotentialVorticity,
]
SURFACE_VARIABLES = [
    met_var.SurfacePressure,
    TOAIncidentSolarRadiation,
    TopNetSolarRadiation,
    TopNetThermalRadiation,
    CloudAreaFraction,
    SurfaceSolarDownwardRadiation,
]

ECMWF_VARIABLES = PRESSURE_LEVEL_VARIABLES + SURFACE_VARIABLES



"""Meteorological, thermophysical, and aircraft constants."""

from __future__ import annotations

# -------
# General
# -------

# NOTE: Use a decimal point for each float-valued constant. This is important for
# converting to numpy arrays.

#: Absolute zero value :math:`[C]`
absolute_zero: float = -273.15

#: Gravitational acceleration :math:`[m \ s^{-2}]`
g: float = 9.80665

#: Radius of Earth :math:`[m]`
radius_earth: float = 6371229.0

#: ISA height of the tropopause :math:`[m]`
#: :cite:`wikipediacontributorsInternationalStandardAtmosphere2023`
h_tropopause: float = 11000.0

# -------------
# Thermodynamic
# -------------

#: Surface pressure, international standard atmosphere :math:`[Pa]`
#: :cite:`wikipediacontributorsInternationalStandardAtmosphere2023`
p_surface: float = 101325.0

#: Isobaric heat capacity of dry air :math:`[J \ kg^{-1} \ K^{-1}]`
c_pd: float = 1004.0  # 1005.7?

#: Isobaric heat capacity of water vapor :math:`[J \ kg^{-1} \ K^{-1}]`
c_pv: float = 1870.0

#: Molecular mass of dry air :math:`[kg \ mol^{-1}]`
M_d: float = 28.9647e-3

#: Molecular mass of water :math:`[kg \ mol^{-1}]`
M_v: float = 18.0153e-3

#: Ratio of heat capacities, TODO: which heat capacities?
gamma: float = 1.4

#:  molar gas constant :math:`[J \ mol^{-1} \ K^{-1}]`
R: float = 8.314462618

#: Gas constant of dry air :math:`[J \ kg^{-1} \ K^{-1}]`
R_d: float = 287.05

#: Gas constant of water vapour :math:`[J \ kg^{-1} \ K^{-1}]`
R_v: float = 461.51

#: Ratio of gas constant for dry air / gas constant for water vapor
epsilon: float = R_d / R_v

#: Density of ice  :math:`[kg \ m^{-3}]`
rho_ice: float = 917.0

#: Adiabatic index air
kappa: float = 1.4

#: Standard atmospheric density at mean sea level (MSL) :math:`[kg \ m^{-3}]`
rho_msl: float = 1.225

#: Standard atmospheric temperature at mean sea level (MSL) :math:`[K]`
T_msl: float = 288.15

#: Speed of sound at mean sea level (MSL) in standard atmosphere :math:`[m \ s^{-1}]`
c_msl: float = 340.294

#: The rate at which the ISA ambient temperature falls with altitude :math:`[K \ m^{-1}]`
#: :cite:`wikipediacontributorsInternationalStandardAtmosphere2023`
T_lapse_rate: float = -0.0065

#: Average incident solar radiation, :math:`[W \ m^{-2}]`
#: This value can range +/- 3% as the earth orbits the sun.
#: From :cite:`UOSRMLSolar` citing :cite:`paltridgeRadiativeProcessesMeteorology1976`
solar_constant: float = 1361.0

# ----
# Fuel
# ----

#: Isobaric heat capacity of combustion products :math:`[J \ kg^{-1} \ K^{-1}]`
c_p_combustion: float = 1250.0

# ------------------
# Optical Properties
# ------------------

#: Real refractive index of ice
mu_ice: float = 1.31

#: Wavelength of visible light (550 nm)
lambda_light: float = 550e-9

#: Ratio between the volume mean radius and the effective radius (Uncertainty +- 0.3)
c_r: float = 0.9

# ------
# Flight
# ------

#: Nominal rate of climb/descent of aircraft [:math:`m \ s^{-1}``]
nominal_rocd: float = 12.7



"""Default CocipGrid parameters."""

from __future__ import annotations

import dataclasses

import numpy as np

from pycontrails.core.fuel import Fuel, JetA

# from pycontrails.models.aircraft_performance import AircraftPerformanceGrid
from pycontrails.models.cocip.cocip_params import CocipParams

try:
    from pycontrails.ext.bada import BADAParams
except ImportError as e:
    raise ImportError(
        'CocipGrid requires BADA extension. Install with `pip install "pycontrails-bada @'
        ' git+ssh://git@github.com/contrailcirrus/pycontrails-bada.git"`'
    ) from e


@dataclasses.dataclass
class CocipGridParams(CocipParams, BADAParams):
    """Default parameters for :class:`CocipGrid`."""

    # ---------
    # Algorithm
    # ---------

    #: Approximate size of a typical `np.array` used with in CoCiP calculations.
    #: The 4-dimensional array defining the waypoint is raveled and split into
    #: batches of this size.
    #: A smaller number for this parameter will reduce memory footprint at the
    #: expense of a longer compute time.
    target_split_size: int = 100_000

    #: Additional boost to target split size before SAC is computed. For typical meshes, only
    #: 10% of waypoints will survive SAC and initial downwash filtering. Accordingly, this parameter
    #: magnifies mesh split size before SAC is computed.
    target_split_size_pre_SAC_boost: float = 3.0

    #: Time slice for loading met interpolators into memory, by default np.timedelta64(1, "h").
    #: Must be a multiple of `np.timedelta64(1, "h")`. The higher the multiple, the more
    #: memory consumed by the method `eval`. If None, the full met dataset is loaded
    #: at once.
    # TODO: Move this to CocipParams once that model is ready for fancy time handling
    met_slice_dt: np.timedelta64 | None = np.timedelta64(1, "h")

    #: Display `tqdm` progress bar showing batch evaluation progress.
    show_progress: bool = True

    # ------------------
    # Simulated Aircraft
    # ------------------

    #: Nominal segment length to place at each grid point [unit `m`]. Round-off error
    #: can be problematic with a small nominal segment length and a large
    #: `dt_integration` parameter. On the other hand, too large of a nominal segment
    #: length diminishes the "locality" of the grid point.
    #: .. versionadded 0.32.2:: EXPERIMENTAL: If None, run CoCiP in "segment-free"
    #: mode. This mode does not include any terms involving segments (wind shear,
    #: segment_length, any derived terms). See :attr:`CocipGridParams.azimuth`
    #: and :attr:`CocipGridParams.dsn_dz_factor` for more details.
    segment_length: float | None = 1000.0

    #: Fuel type
    fuel: Fuel = dataclasses.field(default_factory=JetA)

    #: ICAO code designating simulated aircraft type. Needed to query BADA database.
    aircraft_type: str = "B737"

    #: Engine unique identification number for the ICAO Aircraft Emissions Databank (EDB)
    #: If None, the assumed engine_uid from BADA is used.
    engine_uid: str | None = None

    #: Nominal aircraft wingspan [:math:`m`].
    #: By default, use nominal value derived from BADA.
    wingspan: float | None = None

    #: Navigation bearing [:math:`\deg`] measured in clockwise direction from
    #: true north, by default 0.0.
    #: .. versionadded 0.32.2:: EXPERIMENTAL: If None, run CoCiP in "segment-free"
    #: mode. This mode does not include any terms involving segments (wind shear,
    #: segment_length, any derived terms), unless :attr:`CocipGridParams.dsn_dz_factor`
    #: is non-zero.
    azimuth: float | None = 0.0

    #: Experimental parameter used to approximate ``dsn_dz`` from ``ds_dz`` via
    #: ``dsn_dz = ds_dz * dsn_dz_factor``. A value of 0.0 disables any normal
    #: wind shear effects. An initial unpublished experiment suggests that
    #: ``dsn_dz_factor = 0.665`` adequately approximates the mean EF predictions
    #: of :class:`CocipGrid` over all azimuths.
    #: .. versionadded 0.32.2::
    dsn_dz_factor: float = 0.0

    #: Nominal aircraft mass [:math:`kg`].
    #: By default, use nominal value derived from BADA.
    aircraft_mass: float | None = None

    #: Cruising true airspeed [:math:`m * s^{-1}`].
    #: By default, use nominal value derived from BADA.
    true_airspeed: float | None = None

    #: Nominal engine efficiency, [:math:`0 - 1`].
    #: By default, use values derived from BADA.
    engine_efficiency: float | None = None

    #: Nominal fuel flow, [:math:`kg s^{-1}`].
    #: By default, use values derived from BADA.
    fuel_flow: float | None = None

    #: Nominal thrust, [:math:`N`].
    #: By default, use values derived from BADA.
    thrust: float | None = None

    # #: Aircraft performance model
    # aircraft_performance: AircraftPerformanceGrid | None = None

    # ------------
    # Model output
    # ------------

    #: Attach additional formation specific data to the output. If True, attach
    #: all possible formation data. See :func:`contrail_grid._supported_verbose_outputs_formation`
    #: for a list of supported formation data.
    verbose_outputs_formation: bool | set[str] = False

    #: Attach contrail evolution data to :attr:`CocipGrid.contrail_list`. Requires
    #: substantial memory overhead.
    verbose_outputs_evolution: bool = False



"""Module containing core met variables."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class MetVariable:
    """Met variable defined using CF, ECMWF, and WMO conventions.

    When there is a conflict between CF, ECMWF, and WMO conventions,
    CF takes precendence, then WMO, then ECMWF.

    References
    ----------
    - `CF Standard Names, version 77
    <https://cfconventions.org/Data/cf-standard-names/77/build/cf-standard-name-table.html>`_
    - `ECMWF Parameter Database <https://apps.ecmwf.int/codes/grib/param-db>`_
    - `NCEP Grib v1 Code Table <https://www.nco.ncep.noaa.gov/pmb/docs/on388/table2.html>`
    - `WMO Codes Registry, Grib Edition 2 <https://codes.wmo.int/_grib2>`_
    - `NCEP Grib v2 Code Table <https://www.nco.ncep.noaa.gov/pmb/docs/grib2/grib2_doc/grib2_table4-2.shtml>`

    Used for defining support parameters in a grib-like fashion.
    """  # noqa: E501

    #: Short variable name.
    #: Chosen for greatest consistency between data sources.
    short_name: str

    #: CF standard name, if defined.
    #: Otherwise a standard name is chosen for consistency.
    standard_name: str

    #: Long variable name.
    long_name: str | None = None

    #: Level type
    #: One of "surface", "isobaricInhPa", "nominalTop"
    level_type: str | None = None

    #: ECMWF Grib variable id, if defined.
    #: See `ECMWF Parameter Database <https://apps.ecmwf.int/codes/grib/param-db>`_
    ecmwf_id: int | None = None

    #: WMO Grib v1 variable id, if defined.
    #: See `WMO Codes Registry, Grib Edition 1 <https://codes.wmo.int/_grib1>`_
    #: and `CF Standard Names, version 77 <https://cfconventions.org/Data/cf-standard-names/77/build/cf-standard-name-table.html>`_  # noqa: E501
    grib1_id: int | None = None

    #: WMO Grib 2 variable id, if defined.
    #: See `WMO Codes Registry, Grib Edition 2 <https://codes.wmo.int/_grib2>`_
    #: Tuple represents (disciple, category, number)
    grib2_id: tuple[int, int, int] | None = None

    #: Canonical CF units, if defined.
    units: str | None = None

    #: AMIP identifier, if defined.
    amip: str | None = None

    #: Description
    description: str | None = None

    def __post_init__(self) -> None:
        """Validate inputs.

        Raises
        ------
        ValueError
            If any of the inputs have an unknown :attr:`level_type`.
        """
        level_types = ["surface", "isobaricInPa", "isobaricInhPa", "nominalTop"]
        if self.level_type is not None and self.level_type not in level_types:
            raise ValueError(f"`level_type` must be one of {level_types}")

    @property
    def ecmwf_link(self) -> str | None:
        """Database link in the ECMWF Paramter Database if :attr:`ecmwf_id` is defined.

        Returns
        -------
        str
            Database link in the ECMWF Paramter Database
        """
        return (
            f"https://apps.ecmwf.int/codes/grib/param-db?id={self.ecmwf_id}"
            if self.ecmwf_id
            else None
        )

    @property
    def attrs(self) -> dict[str, str]:
        """Return a dictionary of met variable attributes.

        Compatible with xr.Dataset or xr.DataArray attrs.

        Returns
        -------
        dict[str, str]
            Dictionary with MetVariable attributes.
        """

        # return only these keys if they are not None
        keys = ["short_name", "standard_name", "long_name", "units"]
        return {k: getattr(self, k) for k in keys if getattr(self, k, None) is not None}


# ----
# Dimensions
# ----


AirPressure = MetVariable(
    short_name="p",
    standard_name="air_pressure",
    long_name="Air pressure",
    grib1_id=1,
    ecmwf_id=54,
    units="Pa",
    amip="plev",
    description=(
        "Air pressure is the force per unit area which would be "
        "exerted when the moving gas molecules of which the air is "
        "composed strike a theoretical surface of any orientation."
    ),
)

Altitude = MetVariable(
    short_name="alt",
    standard_name="altitude",
    long_name="Altitude",
    grib1_id=8,
    units="m",
    amip="ta",
    description=(
        "Altitude is the (geometric) height above the geoid, which is the "
        "reference  geopotential surface. The geoid is similar to mean sea level."
    ),
)


# ----
# Pressure level variables
# ----


AirTemperature = MetVariable(
    short_name="t",
    standard_name="air_temperature",
    long_name="Air Temperature",
    units="K",
    level_type="isobaricInhPa",
    grib1_id=11,
    ecmwf_id=130,
    grib2_id=(0, 0, 0),
    amip="ta",
    description=(
        "Air temperature is the bulk temperature of the air, not the surface (skin) temperature."
    ),
)

SpecificHumidity = MetVariable(
    short_name="q",
    standard_name="specific_humidity",
    long_name="Specific Humidity",
    units="kg kg**-1",
    level_type="isobaricInhPa",
    grib1_id=51,
    ecmwf_id=133,
    grib2_id=(0, 1, 0),
    amip="hus",
    description=(
        "Specific means per unit mass. Specific humidity is the mass "
        "fraction of water vapor in (moist) air."
    ),
)

RelativeHumidity = MetVariable(
    short_name="r",
    standard_name="relative_humidity",
    long_name="Relative Humidity",
    units="1",
    level_type="isobaricInhPa",
    grib1_id=52,
    ecmwf_id=157,
    grib2_id=(0, 1, 1),
    amip="hur",
    description=(
        "This parameter is the water vapour pressure as a percentage of "
        "the value at which the air becomes saturated liquid."
    ),
)

Geopotential = MetVariable(
    short_name="z",
    standard_name="geopotential",
    long_name="Geopotential",
    units="m**2 s**-2",
    level_type="isobaricInhPa",
    grib1_id=6,
    ecmwf_id=129,
    grib2_id=(0, 3, 4),
    description=(
        "Geopotential is the sum of the specific gravitational potential energy "
        "relative to the geoid and the specific centripetal potential energy."
    ),
)

GeopotentialHeight = MetVariable(
    short_name="gh",
    standard_name="geopotential_height",
    long_name="Geopotential Height",
    units="m",
    level_type="isobaricInhPa",
    grib1_id=7,
    ecmwf_id=156,
    grib2_id=(0, 3, 5),
    amip="zg",
    description=(
        "Geopotential is the sum of the specific gravitational potential energy "
        "relative to the geoid and the specific centripetal potential energy. "
        "Geopotential height is the geopotential divided by the standard "
        "acceleration due to gravity. It is numerically similar to the altitude "
        "(or geometric height) and not to the quantity with standard name height, "
        "which is relative to the surface."
    ),
)

EastwardWind = MetVariable(
    short_name="u",
    long_name="Eastward Wind",
    standard_name="eastward_wind",
    level_type="isobaricInhPa",
    units="m s**-1",
    grib1_id=33,
    ecmwf_id=131,
    grib2_id=(0, 2, 2),
    amip="ua",
    description=(
        '"Eastward" indicates a vector component which is positive '
        "when directed eastward (negative westward). Wind is defined "
        "as a two-dimensional (horizontal) air velocity vector, with no vertical component."
    ),
)

NorthwardWind = MetVariable(
    short_name="v",
    standard_name="northward_wind",
    long_name="Northward Wind",
    units="m s**-1",
    level_type="isobaricInhPa",
    grib1_id=34,
    ecmwf_id=132,
    grib2_id=(0, 2, 3),
    amip="va",
    description=(
        '"Northward" indicates a vector component which is positive when '
        "directed northward (negative southward). Wind is defined as a "
        "two-dimensional (horizontal) air velocity vector, with no vertical component."
    ),
)

VerticalVelocity = MetVariable(
    short_name="w",
    standard_name="lagrangian_tendency_of_air_pressure",
    long_name="Vertical Velocity (omega)",
    units="Pa s**-1",
    level_type="isobaricInhPa",
    grib1_id=39,
    ecmwf_id=135,
    grib2_id=(0, 2, 8),
    amip="wap",
    description=(
        'The Lagrangian tendency of air pressure, often called "omega", plays '
        "the role of the upward component of air velocity when air pressure "
        "is being used as the vertical coordinate. If the vertical air velocity "
        "is upwards, it is negative when expressed as a tendency of air pressure; "
        "downwards is positive. Air pressure is the force per unit area which "
        "would be exerted when the moving gas molecules of which the air is "
        "composed strike a theoretical surface of any orientation."
    ),
)


# ----
# Single level variables
# ----


SurfacePressure = MetVariable(
    short_name="sp",
    standard_name="surface_air_pressure",
    long_name="Surface air pressure",
    level_type="surface",
    grib1_id=1,
    ecmwf_id=134,
    grib2_id=(0, 3, 0),
    units="Pa",
    amip="ps",
    description=(
        "This parameter is the pressure (force per unit area) of the atmosphere "
        "on the surface of land, sea and in-land water. It is a measure of the "
        "weight of all the air in a column vertically above the area of the "
        "Earth's surface represented at a fixed point."
    ),
)



"""Test interpolation edge cases, etc."""

from __future__ import annotations

import numpy as np
import pytest
import scipy.interpolate

from pycontrails import MetDataArray, MetDataset
from pycontrails.core import interpolation as interp_mod


@pytest.fixture
def mda(met_pcc_pl: MetDataset):
    """Return a MetDataArray for interpolation."""
    assert isinstance(met_pcc_pl, MetDataset)
    return met_pcc_pl["air_temperature"]


def test_basic_interpolation(mda: MetDataArray, caplog):
    """Test basic interpolation patterns."""
    assert mda.shape == (15, 8, 3, 2)

    longitude = np.array([1, 2, 3])
    latitude = np.array([1, 2, 3])
    level = np.array([230, 240, 250])
    t0 = np.datetime64("2019-05-31T05:30")
    time = np.array([t0, t0, t0])

    with caplog.at_level("DEBUG", logger="pycontrails.core.interpolation"):
        out1 = mda.interpolate(longitude, latitude, level, time, bounds_error=True, localize=False)
        assert len(caplog.records) == 0
        out2 = mda.interpolate(longitude, latitude, level, time, bounds_error=True, localize=True)
        # Confirm that the logger in _localize emitted messages
        assert len(caplog.records) == 4

    assert out1.dtype == mda.data.dtype
    assert out2.dtype == mda.data.dtype

    out3 = mda.interpolate(
        longitude.astype("float32"),
        latitude.astype("float32"),
        level.astype("float32"),
        time,
        localize=False,
    )
    out4 = mda.interpolate(
        longitude.astype("float32"),
        latitude.astype("float32"),
        level.astype("float32"),
        time,
        localize=True,
    )
    assert out3.dtype == mda.data.dtype
    assert out4.dtype == mda.data.dtype

    np.testing.assert_array_equal(out1, out2)
    np.testing.assert_array_equal(out3, out4)
    np.testing.assert_allclose(out1, out3, rtol=2e-7)

    for out in [out1, out2, out3, out4]:
        assert np.all(np.isfinite(out))


@pytest.mark.parametrize("localize", [True, False])
@pytest.mark.parametrize("dim", ["longitude", "latitude", "level", "time"])
def test_interpolation_singleton_dim_not_nan(dim: str, mda: MetDataArray, localize: bool):
    """Confirm that interpolation output is not nan when singleton dimension encountered."""
    out_of_bounds_da = mda.data.isel(**{dim: [-1]})
    mda.data = mda.data.isel(**{dim: [0]})

    # Exactly 1 singleton dimension
    assert 1 in mda.shape
    assert mda.shape.count(1) == 1

    coords = {
        "longitude": np.array([1]),
        "latitude": np.array([1]),
        "level": np.array([230]),
        "time": np.array([np.datetime64("2019-05-31T05:30")]),
    }

    # Overwrite coords to include the singleton dimension
    coords[dim] = mda.data[dim].values
    out = mda.interpolate(**coords, bounds_error=True, localize=localize)
    assert np.all(np.isfinite(out))

    # Overwrite coords to include an out of bounds value at the singleton dimension
    coords[dim] = out_of_bounds_da[dim].values
    axis = mda.data.get_axis_num(dim)
    match = f"One of the requested xi is out of bounds in dimension {axis}"
    with pytest.raises(ValueError, match=match):
        mda.interpolate(**coords, bounds_error=True, localize=localize)

    out = mda.interpolate(**coords, bounds_error=False, localize=localize)
    np.testing.assert_array_equal(out, [np.nan])

    out = mda.interpolate(**coords, bounds_error=False, localize=localize, fill_value=999)
    np.testing.assert_array_equal(out, [999])


@pytest.mark.parametrize("localize", [True, False])
def test_interpolation_single_level(mda: MetDataArray, localize: bool):
    """Confirm interpolation works with single level (level = -1) DataArray."""
    # Convert mda to a DataArray with a single level
    mda.data = mda.data.isel(level=[0]).assign_coords(level=[-1.0])
    np.testing.assert_array_equal(mda.data.level, [-1])
    assert mda.shape == (15, 8, 1, 2)

    longitude = [30, 40, 200]
    latitude = [30, 40, 60]
    level = [300, 400, 500]
    time = [
        np.datetime64("2019-05-31T05:30"),
        np.datetime64("2019-05-31T05:33"),
        np.datetime64("2019-05-31T05:36"),
    ]

    # Longitude is out of bounds
    with pytest.raises(ValueError, match="One of the requested xi is out of bounds in dimension 0"):
        mda.interpolate(longitude, latitude, level, time, bounds_error=True, localize=localize)

    # Only the third coordinate is out of bounds .... level interpolation works as intended
    out = mda.interpolate(longitude, latitude, level, time, bounds_error=False, localize=localize)
    np.testing.assert_array_equal(np.isnan(out), [False, False, True])


def test_localize(mda: MetDataArray):
    """Confirm _localize implementation."""
    coords = {}
    coords["longitude"] = np.array([1, 2, 3])
    coords["latitude"] = np.array([1, 2, 3])
    coords["level"] = np.array([250, 250, 250])
    t0 = np.datetime64("2019-05-31T05:30")
    coords["time"] = np.array([t0, t0, t0])

    da = mda.data
    out = interp_mod._localize(da, coords)
    assert out.ndim == 4
    np.testing.assert_array_equal(out["longitude"], [0, 25])
    np.testing.assert_array_equal(out["latitude"], [-10, 15])
    np.testing.assert_array_equal(out["level"], [250])
    np.testing.assert_array_equal(out["time"], da["time"])

    # Try one more situation
    coords["longitude"] = np.array([-180, -180, -190])
    coords["latitude"] = np.array([-100, 85, 86])
    coords["level"] = np.array([230, 240, 220])

    out = interp_mod._localize(da, coords)
    assert out.ndim == 4
    np.testing.assert_array_equal(out["longitude"], [-160])
    np.testing.assert_array_equal(out["latitude"], da["latitude"])
    np.testing.assert_array_equal(out["level"], [225.0, 250.0])
    np.testing.assert_array_equal(out["time"], da["time"])

    # And another situation
    t0 = np.datetime64("2019-05-31T06:00")
    coords["time"] = np.array([t0, t0, t0])
    out = interp_mod._localize(da, coords)
    assert out.ndim == 4
    np.testing.assert_array_equal(out["longitude"], [-160])
    np.testing.assert_array_equal(out["latitude"], da["latitude"])
    np.testing.assert_array_equal(out["level"], [225.0, 250.0])
    np.testing.assert_array_equal(out["time"], [t0])


@pytest.mark.parametrize("localize", [True, False])
def test_interpolation_time_resolution(mda: MetDataArray, localize: bool):
    """Confirm interpolation works as expected with time to float conversions."""
    longitude = 30
    latitude = 31
    level = 234
    time = np.datetime64("2019-05-31T06:00:01")

    # Time is out of bounds
    match = "One of the requested xi is out of bounds in dimension 3"
    with pytest.raises(ValueError, match=match):
        mda.interpolate(longitude, latitude, level, time, bounds_error=True, localize=localize)

    # If time is identical with the endpoint of the time dimension,
    # it should be in bounds.
    time = np.datetime64("2019-05-31T06:00:00")

    out = mda.interpolate(longitude, latitude, level, time, bounds_error=False, localize=localize)
    assert np.isfinite(out)

    # Increasing it by 1 us makes it out of bounds
    time += np.timedelta64(1, "us")
    out = mda.interpolate(longitude, latitude, level, time, bounds_error=False, localize=localize)
    assert np.isnan(out)


@pytest.mark.parametrize("dtype", ["float32", "float64"])
def test_regular_4d_grid_interpolator(mda: MetDataArray, dtype: str):
    """Confirm ``PycontrailsRegularGridInterpolator`` agrees with scipy parent class."""
    x = mda.data["longitude"].values
    y = mda.data["latitude"].values
    z = mda.data["level"].values
    t = mda.data["time"].values
    t = interp_mod._floatize_time(t, t[0])
    assert x.dtype == np.float64
    assert y.dtype == np.float64
    assert z.dtype == np.float64
    assert t.dtype == np.float64

    mda.data = mda.data.astype(dtype)

    points = x, y, z, t
    values = mda.values

    kwargs = {"method": "linear", "bounds_error": True, "fill_value": np.nan}
    rgi1 = interp_mod.PycontrailsRegularGridInterpolator(points, values, **kwargs)
    rgi2 = scipy.interpolate.RegularGridInterpolator(points, values, **kwargs)

    rng = np.random.default_rng(6567)
    n = 100_000
    x0 = rng.uniform(x.min(), x.max(), size=n)
    y0 = rng.uniform(y.min(), y.max(), size=n)
    z0 = rng.uniform(z.min(), z.max(), size=n)
    t0 = rng.uniform(t.min(), t.max(), size=n)
    xi = np.stack([x0, y0, z0, t0], axis=1)

    out1 = rgi1(xi)
    out2 = rgi2(xi)

    # The pycontrails version always uses the same dtype as the underlying values array
    assert out1.dtype == mda.data.dtype

    # The scipy version essentially always promotes to float64
    assert out2.dtype == np.float64

    np.testing.assert_allclose(out1, out2, rtol=1e-7)


@pytest.mark.parametrize("localize", [True, False])
@pytest.mark.parametrize("dim", ["longitude", "latitude", "level", "time"])
def test_regular_3d_grid_interpolator(mda: MetDataArray, dim: str, localize: bool):
    """Confirm ``PycontrailsRegularGridInterpolator`` can handle singleton dimensions.

    Previously, scipy 1.8 could not handle singleton dimensions. This was fixed in
    scipy 1.9 and the pycontrails homegrown implementation was removed in
    pycontrails 0.25.6.

    Implementations changed again in scipy 1.10 and pycontrails ~0.35.
    """
    da = mda.data.isel(**{dim: [0]})
    x = da["longitude"].values
    y = da["latitude"].values
    z = da["level"].values
    t = da["time"].values

    rng = np.random.default_rng(6567)
    n = 100_000
    x0 = rng.uniform(x.min(), x.max(), size=n)
    y0 = rng.uniform(y.min(), y.max(), size=n)
    z0 = rng.uniform(z.min(), z.max(), size=n)
    t0 = rng.uniform(t.min(), t.max(), size=n).astype("datetime64[ns]")

    # Run interpolation through pycontrails interface
    kwargs = {"method": "linear", "bounds_error": True, "fill_value": np.nan}
    out1 = interp_mod.interp(x0, y0, z0, t0, da=da, localize=localize, **kwargs)
    assert np.all(np.isfinite(out1))

    # Run interpolation through scipy RGI
    # First "floatize" time (this is done automatically in interp_mod.interp)
    t_float = interp_mod._floatize_time(t, t[0])
    t0 = interp_mod._floatize_time(t0, t[0])

    points = x, y, z, t_float
    values = da.values
    rgi = scipy.interpolate.RegularGridInterpolator(points, values, **kwargs)
    xi = np.stack([x0, y0, z0, t0], axis=1)

    # We no longer get nan values with single dimension interpolation
    # This was fixed in scipy 1.9.0. All values are finite now.
    out2 = rgi(xi)
    assert np.all(np.isfinite(out2))

    np.testing.assert_allclose(out1, out2, rtol=1e-7)


@pytest.mark.parametrize("method", ["nearest", "slinear", "cubic", "quintic"])
def test_scipy19_interpolation_methods(mda: MetDataArray, method: str):
    """Confirm ``PycontrailsRegularGridInterpolator`` can handle methods introduced in scipy 1.9."""
    da = mda.data
    assert da.shape == (15, 8, 3, 2)
    assert da.size == 720

    x = da["longitude"].values
    y = da["latitude"].values
    z = da["level"].values
    t = da["time"].values

    rng = np.random.default_rng(4345)
    x0 = rng.uniform(x.min(), x.max(), size=1000)
    y0 = rng.uniform(y.min(), y.max(), size=1000)
    z0 = rng.uniform(z.min(), z.max(), size=1000)
    t0 = rng.uniform(t.min(), t.max(), size=1000).astype("datetime64[ns]")

    # Run interpolation through pycontrails interface
    kwargs = {"method": method, "bounds_error": True, "fill_value": np.nan, "localize": True}
    if method in ["cubic", "quintic"]:
        with pytest.raises(ValueError, match="The number of derivatives at boundaries does not"):
            interp_mod.interp(x0, y0, z0, t0, da=da, **kwargs)
        return

    out1 = interp_mod.interp(x0, y0, z0, t0, da=da, **kwargs)
    assert np.all(np.isfinite(out1))
    assert out1.dtype == "float32"

    kwargs = {"method": "linear", "bounds_error": True, "fill_value": np.nan, "localize": True}
    out2 = interp_mod.interp(x0, y0, z0, t0, da=da, **kwargs)
    assert np.all(np.isfinite(out2))

    if method == "slinear":
        np.testing.assert_array_equal(out1, out2)
        return

    # Pin the RMSE (out of curiosity)
    rmse = (np.mean((out1 - out2) ** 2)) ** 0.5
    assert rmse == pytest.approx(3.3319390877860893, rel=1e-10)


@pytest.fixture
def arbitrary_coords():
    longitude = np.array([-55, -43, -55, -21, -17, -18], dtype=np.float32)
    latitude = np.array([12, 17, -8, -44, 22, 23], dtype=np.float32)
    level = np.array([227, 228, 231, 233, 231, 230], dtype=np.float32)
    time = np.array(
        [
            np.datetime64("2019-05-31T05:30"),
            np.datetime64("2019-05-31T05:33"),
            np.datetime64("2019-05-31T05:36"),
            np.datetime64("2019-05-31T05:40"),
            np.datetime64("2019-05-31T05:43"),
            np.datetime64("2019-05-31T05:46"),
        ]
    )
    return longitude, latitude, level, time


def test_indices_distinct_vars(
    met_pcc_pl: MetDataset,
    arbitrary_coords: tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
):
    """Check that indices are consistent when interpolating over distinct variables."""

    assert list(met_pcc_pl) == [
        "air_temperature",
        "specific_humidity",
        "specific_cloud_ice_water_content",
    ]

    out1, indices1 = interp_mod.interp(
        *arbitrary_coords,
        met_pcc_pl["air_temperature"].data,
        method="linear",
        bounds_error=True,
        fill_value=np.nan,
        localize=False,
        return_indices=True,
    )
    assert np.all(np.isfinite(out1))

    out2, indices2 = interp_mod.interp(
        *arbitrary_coords,
        met_pcc_pl["specific_humidity"].data,
        method="linear",
        bounds_error=True,
        fill_value=np.nan,
        localize=False,
        return_indices=True,
    )
    assert np.all(np.isfinite(out2))

    np.testing.assert_array_equal(indices1.xi_indices, indices2.xi_indices)
    np.testing.assert_array_equal(indices1.norm_distances, indices2.norm_distances)
    assert indices1.out_of_bounds is None
    assert indices2.out_of_bounds is None


@pytest.mark.parametrize(
    "var", ["air_temperature", "specific_humidity", "specific_cloud_ice_water_content"]
)
def test_indices_same_var(
    met_pcc_pl: MetDataset,
    arbitrary_coords: tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
    var: str,
):
    """Check that interpolation with and without indices gives the same result."""

    da = met_pcc_pl[var].data

    out1, indices = interp_mod.interp(
        *arbitrary_coords,
        da,
        method="linear",
        bounds_error=True,
        fill_value=np.nan,
        localize=False,
        return_indices=True,
    )

    # Expect this to be faster because indices are already computed
    out2 = interp_mod.interp(
        *arbitrary_coords,
        da,
        method="linear",
        bounds_error=True,
        fill_value=np.nan,
        localize=False,
        indices=indices,
        return_indices=False,
    )
    np.testing.assert_array_equal(out1, out2)


@pytest.mark.parametrize("bounds_error", [False, True])
@pytest.mark.parametrize("idx", range(5))
@pytest.mark.parametrize("coord", ["longitude", "latitude", "level", "time"])
def test_interpolation_propagate_nan(mda: MetDataArray, bounds_error: bool, idx: int, coord: str):
    """Ensure nan values propagate through interpolation."""

    longitude = np.arange(5, dtype=float)
    latitude = np.arange(5, 10, dtype=float)
    level = np.linspace(225, 250, 5)
    time = np.r_[[np.datetime64("2019-05-31T05:20")] * 3, [np.datetime64("2019-05-31T05:36")] * 2]

    if coord == "longitude":
        longitude[idx] = np.nan
        dim = 0
    elif coord == "latitude":
        latitude[idx] = np.nan
        dim = 1
    elif coord == "level":
        level[idx] = np.nan
        dim = 2
    elif coord == "time":
        time[idx] = np.datetime64("NaT")
        dim = 3

    if bounds_error:
        match = f"One of the requested xi is out of bounds in dimension {dim}"
        with pytest.raises(ValueError, match=match):
            mda.interpolate(longitude, latitude, level, time, bounds_error=bounds_error)
        return

    out = mda.interpolate(longitude, latitude, level, time, bounds_error=bounds_error)
    assert np.flatnonzero(np.isnan(out)).item() == idx


@pytest.mark.parametrize("replace_value", [np.nan, 224])
@pytest.mark.parametrize("fill_value", [np.nan, 0.0, None, 55.0])
def test_fill_value(
    mda: MetDataArray,
    arbitrary_coords: tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
    replace_value: float | np.float64,
    fill_value: float | np.float64 | None,
):
    """Check implementation with nonstandard fill_value."""

    da = mda.data
    assert np.isnan(replace_value) or replace_value < da.level.min()
    arbitrary_coords[2][3] = replace_value

    with pytest.raises(ValueError, match="One of the requested xi is out of bounds in dimension 2"):
        interp_mod.interp(
            *arbitrary_coords,
            da,
            method="linear",
            bounds_error=True,
            fill_value=fill_value,
            localize=False,
        )

    out1, indices = interp_mod.interp(
        *arbitrary_coords,
        da,
        method="linear",
        bounds_error=False,
        fill_value=fill_value,
        localize=False,
        return_indices=True,
    )

    out2 = interp_mod.interp(
        *arbitrary_coords,
        da,
        method="linear",
        bounds_error=False,
        fill_value=fill_value,
        localize=False,
        indices=indices,
    )

    np.testing.assert_array_equal(out1, out2)
    if np.isnan(replace_value):
        assert np.isnan(out1[3])
        return
    if fill_value is None:
        assert out1[3] == pytest.approx(214.0025, abs=1e-4)
        return
    if np.isnan(fill_value):
        assert np.isnan(out1[3])
        return
    assert np.isfinite(out1[3])
    assert out1[3] == fill_value



"""GFS Data Access."""

from __future__ import annotations

from pycontrails.datalib.gfs.gfs import GFS_FORECAST_BUCKET, GFSForecast
from pycontrails.datalib.gfs.variables import (
    GFS_VARIABLES,
    PRESSURE_LEVEL_VARIABLES,
    SURFACE_VARIABLES,
    CloudIceWaterMixingRatio,
    TOAUpwardLongwaveRadiation,
    TOAUpwardShortwaveRadiation,
    TotalCloudCoverIsobaric,
    Visibility,
)

__all__ = [
    "GFS_FORECAST_BUCKET",
    "GFSForecast",
    "CloudIceWaterMixingRatio",
    "TotalCloudCoverIsobaric",
    "Visibility",
    "TOAUpwardShortwaveRadiation",
    "TOAUpwardLongwaveRadiation",
    "GFS_VARIABLES",
    "PRESSURE_LEVEL_VARIABLES",
    "SURFACE_VARIABLES",
]
